/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023,2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include <librnd/config.h>
#include <librnd/hid/hid.h>
#include <librnd/hid/hid_inlines.h>

#include "conf_core.h"
#include "obj_common.h"
#include "obj_way.h"
#include "obj_any.h"
#include "draw.h"
#include "geo.h"
#include "projection.h"

static void osm_way_free_fields(osm_any_obj_t *obj)
{
	osm_tags_uninit(&obj->way.curr->tags);
	vtI0_uninit(&obj->way.curr->nd);
	vtp0_uninit(&obj->way.ndp);
	obj->way.ndp_rev = -1;

	if (obj->way.curr != &obj->way.orig) {
		free(obj->way.curr);
		obj->way.curr = &obj->way.orig;
	}
}

RND_INLINE void osm_way_bbox_(osm_any_obj_t *obj)
{
	long n;

	if (obj->way.ndp.used > 0) {
		obj->way.hdr.bbox = *osm_obj_bbox(obj->way.ndp.array[0]);
		for(n = 1; n < obj->way.ndp.used; n++)
			osm_rtree_box_bump(&obj->way.hdr.bbox, osm_obj_bbox(obj->way.ndp.array[n]));
	}

	obj->way.hdr.bbox_valid = 1;

}

/* Re-generate the ndp cache if the layer has changed since last generated */
RND_INLINE void osm_way_ensure_ndp_(osm_way_t *way)
{
	long n;
	osm_layer_t *ly = way->hdr.layer;

	if (!way->hdr.dirty && (way->ndp_rev >= ly->rev))
		return;

	/* need to re-generate the cache */
	way->ndp.used = 0;
	for(n = 0; n < way->curr->nd.used; n++) {
		osm_node_t *nd = htIp_get(&ly->id2obj[OSM_OBJ_NODE], way->curr->nd.array[n]);
		if (nd == NULL)
			rnd_message(RND_MSG_ERROR, "osm_way_ensure_ndp_(): invalid id %lu\n", (unsigned long)way->curr->nd.array[n]);
		else
			vtp0_append(&way->ndp, nd);
	}

	way->ndp_rev = ly->rev;

	if (way->hdr.dirty) {
		osm_any_obj_t *o = (osm_any_obj_t *)way;
		osm_objcom_rtree_unreg(o);
		o->way.hdr.bbox_valid = 0;
		osm_way_bbox_(o);
		osm_objcom_rtree_reg(o);

		way->hdr.dirty = 0;
	}
}

void osm_way_ensure_ndp(osm_way_t *way)
{
	osm_way_ensure_ndp_(way);
}


typedef enum dir_ind_e { /* bitfield */
	DIR_IND_NONE  = 0,
	DIR_IND_MID   = 1,
	DIR_IND_START = 2,
	DIR_IND_END   = 4
} dir_ind_t;

RND_INLINE void osm_way_draw_line(rnd_hid_gc_t gc, rnd_hid_gc_t gc_dir, rnd_coord_t x1, rnd_coord_t y1, rnd_coord_t x2, rnd_coord_t y2, dir_ind_t di, double dir_size)
{
	double dx, dy, nx, ny, len;
	rnd_coord_t mx, my;

	rnd_render->draw_line(gc, x1, y1, x2, y2);

	if (di == DIR_IND_NONE)
		return;

	mx = (x1+x2)/2; my = (y1+y2)/2;
	dx = x2 - x1; dy = y2 - y1;
	len = dx*dx+dy*dy;
	if (len == 0)
		return;
	len = sqrt(len);

	if (len < dir_size*4)
		return;

	dx /= len; dy /= len;
	nx = -dy; ny = dx;

	rnd_render->draw_line(gc_dir, mx, my, mx - nx*dir_size - dx*dir_size, my - ny*dir_size - dy*dir_size);
	rnd_render->draw_line(gc_dir, mx, my, mx + nx*dir_size - dx*dir_size, my + ny*dir_size - dy*dir_size);

	if (di & DIR_IND_START)
		rnd_render->draw_line(gc_dir, mx + nx*dir_size - dx*dir_size, my + ny*dir_size - dy*dir_size, mx - nx*dir_size - dx*dir_size, my - ny*dir_size - dy*dir_size);

	if (di & DIR_IND_END)
		rnd_render->draw_line(gc_dir, mx + nx*dir_size, my + ny*dir_size, mx - nx*dir_size, my - ny*dir_size);
}

/* Low level draw: assumes gc is set up to style */
RND_INLINE void osm_way_draw_(osm_any_obj_t *obj, rnd_hid_gc_t gc, rnd_hid_gc_t gc_dir, int att, double dir_size)
{
	long n;
	osm_node_t *prev, *curr;
	int want_dir = (dir_size != 0);

	osm_way_ensure_ndp_(&obj->way);

	prev = obj->way.ndp.array[0];
	for(n = 1; n < obj->way.ndp.used; n++) {
		dir_ind_t di = DIR_IND_NONE;

		curr = obj->way.ndp.array[n];
		if (want_dir) {
			di = DIR_IND_MID;
			if (n == 1) di |= DIR_IND_START;
			if (n == (obj->way.ndp.used-1)) di |= DIR_IND_END;
		}

		if (att)
			osm_way_draw_line(gc, gc_dir, prev->att_x, prev->att_y, curr->att_x, curr->att_y, di, dir_size);
		else
			osm_way_draw_line(gc, gc_dir, prev->x, prev->y, curr->x, curr->y, di, dir_size);
		prev = curr;
	}
}

/* gc used to indicate direction (cached as long as HID doesn't change) */
static rnd_hid_gc_t gc_dir;
static rnd_hid_t *gc_dir_hid = NULL;

static void osm_way_draw(osm_any_obj_t *obj, rnd_hid_gc_t gc, const rnd_hid_expose_ctx_t *region, rnd_xform_t *xform, int overlay)
{

	if (gc_dir_hid != rnd_render) {
		if (gc_dir_hid != NULL)
			rnd_render->destroy_gc(gc_dir);
		
		gc_dir_hid = rnd_render;
		if (rnd_render != NULL) {
			gc_dir = rnd_render->make_gc(rnd_render);
			rnd_hid_set_line_cap(gc_dir, rnd_cap_round);
			rnd_hid_set_line_width(gc_dir, -2);
			rnd_render->set_color(gc_dir, rnd_color_magenta);
		}
	}

	if (obj->hdr.style == NULL) {
		rnd_hid_set_line_cap(gc, rnd_cap_round);
		rnd_hid_set_line_width(gc, -1);
		rnd_render->set_color(gc, rnd_color_magenta);
	}
	else {
		osm_style_data_t *sd;

		if (!obj->hdr.selected && (xform->zoom_mpp > obj->hdr.style->lod_invis_above))
			return;
		else if (obj->hdr.selected || (xform->zoom_mpp < obj->hdr.style->lod_high_below))
			sd = &obj->hdr.style->high;
		else
			sd = &obj->hdr.style->low;

		rnd_hid_set_line_cap(gc, rnd_cap_round);
		rnd_hid_set_line_width(gc, conf_core.editor.thin_draw ? -1 : sd->stroke.width);
		rnd_render->set_color(gc, obj->hdr.selected ? &conf_core.appearance.color.selected : &sd->stroke.color);
	}
	osm_way_draw_(obj, gc, gc_dir, 0, obj->hdr.selected ? 800 * xform->zoom_mpp: 0);
}

static void osm_way_draw_attached(osm_any_obj_t *obj, rnd_hid_gc_t gc)
{
	osm_way_draw_(obj, gc, gc_dir, 1, 0);
}


static void osm_way_bbox(osm_any_obj_t *obj)
{
	osm_way_ensure_ndp_(&obj->way);
	osm_way_bbox_(obj);
}

static osm_any_obj_t *osm_way_alloc(void) { return (osm_any_obj_t *)osm_way_new(); }

static int osm_way_isc_box(const osm_any_obj_t *o, const osm_rtree_box_t *box)
{
	long n;
	osm_node_t *prev, *curr;

	if (o->way.ndp.used < 2)
		return 0;

	prev = o->way.ndp.array[0];
	for(n = 1; n < o->way.ndp.used; n++) {
		int isc;

		curr = o->way.ndp.array[n];

		isc = osm_isc_line_rectangle(
			box->x1, box->y1, box->x2, box->y2,
			prev->x, prev->y, curr->x, curr->y);

		if (isc)
			return 1;

		prev = curr;
	}
	return 0;
}

static osm_tags_t *osm_way_get_tags(osm_any_obj_t *o)
{
	return &o->way.curr->tags;
}

void osm_way_data_copy(osm_way_data_t *dst, const osm_way_data_t *src, int deep)
{
	if (deep) {
		long n;

		vtI0_init(&dst->nd);
		vtI0_enlarge(&dst->nd, src->nd.used-1);
		for(n = 0; n < src->nd.used; n++)
			dst->nd.array[n] = src->nd.array[n];

		osm_tags_dup(&dst->tags, &src->tags);
	}
}

static void osm_way_fork(osm_way_t *way)
{
	assert(way->curr == &way->orig);

	way->curr = malloc(sizeof(osm_way_data_t));
	osm_way_data_copy(way->curr, &way->orig, 1);
}

static void osm_way_unfork(osm_way_t *way)
{
	assert(way->curr != &way->orig);
	vtI0_uninit(&way->curr->nd);
	osm_tags_uninit(&way->curr->tags);
	free(way->curr);
	way->curr = &way->orig;
}

void osm_way_pre_edit(osm_any_obj_t *o, int geo, int auto_fork)
{
	if (auto_fork && (o->way.curr == &o->way.orig))
		osm_way_fork(&o->way);
	if (geo) {
		osm_objcom_rtree_unreg(o);
		o->way.hdr.bbox_valid = 0;
	}
}


void osm_way_post_edit(osm_any_obj_t *o, int tags_changed)
{
	o->way.ndp_rev = -1;
	if (tags_changed)
		osm_style_update_obj(o);
	if (!o->way.hdr.bbox_valid) {
		osm_way_bbox(o);
		osm_objcom_rtree_reg(o);
	}
	o->hdr.layer->rev++;
}

static void osm_way_referee_nodes(osm_any_obj_t *obj, vtp0_t *dst, int update_att)
{
	long n;

	osm_way_ensure_ndp_(&obj->way);

	for(n = 0; n < obj->way.ndp.used; n++)
		osm_node_referee_nodes(obj->way.ndp.array[n], dst, update_att);
}

static long osm_way_referee_idx(osm_any_obj_t *parent, osm_any_obj_t *child)
{
	long n;

	osm_way_ensure_ndp_(&parent->way);

	for(n = 0; n < parent->way.ndp.used; n++)
		if (parent->way.ndp.array[n] == child)
			return n;

	return -1;
}

static void osm_way_diffresult(osm_any_obj_t *o, osm_id_t new_id, osm_id_t new_version)
{
	osm_tags_uninit(&o->way.orig.tags);
	vtI0_uninit(&o->way.orig.nd);
	osm_way_data_copy(&o->way.orig, o->way.curr, 0);
	o->way.orig.tags = o->way.curr->tags;
	o->way.orig.nd = o->way.curr->nd;
	free(o->way.curr);
	o->way.curr = &o->way.orig;
	osm_objcom_diffresult(o, new_id, new_version);
}

static void osm_way_diffresult_referees(osm_any_obj_t *o, htdr_t *dr, int permit_del, int undoable)
{
	long n;
	htdr_key_t key;

	key.type = OSM_OBJ_NODE;

	for(n = 0; n < o->way.curr->nd.used; n++) {
		htdr_value_t val;

		key.old_id = o->way.curr->nd.array[n];
		val = htdr_get(dr, key);
		if (val.valid) {
			if (val.new_id == 0) {
				rnd_message(RND_MSG_ERROR, "osm_way_diffresult_referees(): unexpected delete\n");
				continue;
			}
			if (undoable) {
				int collapsed = 0;
				if ((n > 0) && (o->way.curr->nd.array[n-1] == val.new_id)) {
					osm_way_modify_remove(&o->way, n, 1);
					collapsed = 1;
				}
				else if ((n < o->way.curr->nd.used-1) && (o->way.curr->nd.array[n+1] == val.new_id)) {
					osm_way_modify_remove(&o->way, n, 1);
					collapsed = 1;
				}
				else
					osm_way_modify_replace(&o->way, n, val.new_id, 1);

				while (collapsed) {
					/* [n] got removed, now check if its original neighbors
					   n[-1] and n[+1] are the same */
					collapsed = 0;
					if ((n > 0) && (n < o->way.curr->nd.used) && (o->way.curr->nd.array[n] == o->way.curr->nd.array[n-1])) {
						osm_way_modify_remove(&o->way, n, 1);
						collapsed = 1;
					}
				}
			}
			else
				o->way.curr->nd.array[n] = val.new_id;
		}
	}
}

static void osm_way_pre_remove_referee(osm_any_obj_t *o, osm_objtype_t rem_ty, osm_id_t rem_id, int undoable)
{
	long n, idx = -1;
	int close_again = 0;

	if (rem_ty != OSM_OBJ_NODE)
		return; /* ways refer to nodes only */

	/* find index */
	for(n = 0; n < o->way.curr->nd.used; n++) {
		if (o->way.curr->nd.array[n] == rem_id) {
			if (n == 0) {
				/* check for special case: closed loop: first node is the same as last node */
				if (o->way.curr->nd.array[o->way.curr->nd.used-1] == rem_id)
					close_again = 1;
			}
			idx = n;
			osm_way_modify_remove(&o->way, idx, undoable);
			n--;
		}
	}

	if (idx < 0)
		return;

	if (o->way.ndp.used < 2) {
		osm_obj_remove(o, undoable);
		return;
	}

	if (close_again)
		osm_way_modify_append(o->hdr.layer, &o->way, o->way.curr->nd.array[0], undoable);
}

static void osm_way_post_remove(osm_any_obj_t *o, int undoable)
{
	long n;
	/* remove orphaned nodes */
	for(n = o->way.ndp.used-1; n >= 0; n--) {
		osm_node_t *node = o->way.ndp.array[n];
		if ((node->hdr.layer != NULL) && (osm_node_is_orphan(node, o)))
			osm_obj_remove((osm_any_obj_t *)node, undoable);
	}
}

osm_any_obj_t *osm_way_dup_into(osm_any_obj_t *o, osm_layer_t *dst, osm_id_t new_id)
{
	osm_any_obj_t *obj = osm_way_alloc();
	osm_way_data_copy(obj->way.curr, o->way.curr, 1);
	obj->way.ndp_rev = -1;
	if (new_id != 0)
		obj->hdr.id = new_id;
	else
		obj->hdr.id = o->hdr.id;
	osm_obj_add_to_layer(dst, obj);
	return obj;
}

static const osm_objcalls_t osm_way_calls = {
	osm_way_alloc,
	osm_way_free_fields,
	osm_way_draw,
	osm_way_draw_attached,
	osm_way_bbox,
	osm_way_isc_box,
	osm_way_get_tags,
	osm_way_pre_edit,
	osm_way_post_edit,
	osm_way_referee_nodes,
	osm_way_referee_idx,
	osm_way_diffresult,
	osm_way_diffresult_referees,
	osm_way_pre_remove_referee,
	osm_way_post_remove,
	osm_way_dup_into
};

void osm_way_init(osm_way_t *way)
{
	memset(way, 0, sizeof(osm_way_t));
	osm_objcom_init(&way->hdr, OSM_OBJ_WAY, &osm_way_calls);
	way->curr = &way->orig;
	osm_tags_init(&way->orig.tags);
	way->ndp_rev = -1;
}

osm_way_t *osm_way_new(void)
{
	osm_way_t *res = malloc(sizeof(osm_way_t));
	osm_way_init(res);
	return res;
}

int osm_way_refers_node(osm_way_t *way, osm_node_t *node)
{
	long n;

	osm_way_ensure_ndp_(way);

	for(n = 0; n < way->ndp.used; n++)
		if (way->ndp.array[n] == node)
			return 1;

	return 0;
}

long osm_way_list_referers_by_node(vtp0_t *lst, osm_node_t *node)
{
	osm_rtree_it_t it;
	void *o;
	long count = 0;

	for(o = osm_rtree_first(&it, &node->hdr.layer->objs, &node->hdr.bbox); o != NULL; o = osm_rtree_next(&it)) {
		osm_way_t *way = o;
		if ((way->hdr.type == OSM_OBJ_WAY) && !way->hdr.referer && osm_way_refers_node(way, node)) {
			vtp0_append(lst, way);
			way->hdr.referer = 1;
			count++;
		}
	}
	return count;
}

long osm_way_dirt_referers_by_node(osm_node_t *node)
{
	osm_rtree_it_t it;
	void *o;
	long count = 0;

	for(o = osm_rtree_first(&it, &node->hdr.layer->objs, &node->hdr.bbox); o != NULL; o = osm_rtree_next(&it)) {
		osm_way_t *way = o;
		if ((way->hdr.type == OSM_OBJ_WAY) && !way->hdr.referer && osm_way_refers_node(way, node)) {
			way->hdr.dirty = 1;
			count++;
		}
	}
	return count;
}

long osm_way_find_near(osm_way_t *way, rnd_coord_t px, rnd_coord_t py, rnd_coord_t maxdist2, rnd_coord_t *best_x_out, rnd_coord_t *best_y_out)
{
	long n, best_idx = -1;
	osm_node_t *prev, *curr;
	double best_dist2 = maxdist2;
	rnd_coord_t best_x, best_y;

	if (way->ndp.used < 2)
		return -1;

	osm_way_ensure_ndp_(way);

	prev = way->ndp.array[0];
	for(n = 1; n < way->ndp.used; n++,prev = curr) {
		double dist2, dx, dy, offs;
		rnd_coord_t lx, ly;
		curr = way->ndp.array[n];

		offs = osm_proj_pt_onto_line(&lx, &ly, px, py, prev->x, prev->y, curr->x, curr->y);
		if ((offs < 0.0) || (offs > 1.0))
			continue; /* projection is outside of the segment */

		dx = lx - px; dy = ly - py;
		dist2 = dx*dx + dy*dy;

		if (dist2 <= best_dist2) {
			best_x = lx; best_y = ly;
			best_idx = n;
		}
	}

	*best_x_out = best_x;
	*best_y_out = best_y;
	return best_idx;
}

#include "obj_way_undo.c"
