/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - Bing map tiles
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"
#include <math.h>
#include <librnd/config.h>
#include <librnd/core/hidlib.h>
#include <librnd/core/conf_multi.h>
#include <librnd/core/actions.h>
#include <librnd/hid/hid.h>
#include <librnd/hid/hid_inlines.h>
#include <librnd/hid/pixmap.h>

#include "data.h"
#include "event.h"
#include "projection.h"

#include "img_bing_conf.h"

conf_bing_t conf_bing;

static const char img_bing_cookie[] = "img_bing";

#include "conf_internal.c"

/*** Coordinate conversions: Bing pixels/tiles/quadkey; see: ***/
/* See: https://learn.microsoft.com/en-us/bingmaps/articles/bing-maps-tile-system */

RND_INLINE double clip(double val, double min, double max)
{
	if (val < min) return min;
	if (val > max) return max;
	return val;
}

/* Convert lon/lat degs to Bing pixel x;y */
RND_INLINE void bing_latlon2xy(double lon, double lat, int zoom, long *px, long *py)
{
	static const double min_lon = -180, max_lon = 180;
	static const double min_lat = -85.05112878, max_lat = 85.05112878;
	double x, y, sinlat;
	unsigned long map_size = 256UL << zoom;

	lat = clip(lat, min_lat, max_lat);
	lon = clip(lon, min_lon, max_lon);

	sinlat = sin(lat * M_PI / 180.0);
	x = (lon + 180.0) / 360.0;
	y = 0.5 - log((1 + sinlat) / (1 - sinlat)) / (4.0 * M_PI);

	*px = clip(x * map_size + 0.5, 0, map_size - 1);
	*py = clip(y * map_size + 0.5, 0, map_size - 1);
}

/* Convert Bing pixel px;py to lon/lat degs */
RND_INLINE void bing_xy2latlon(long px, long py, int zoom, double *lon, double *lat)  
{
	unsigned long map_size = 256UL << zoom;
	double x = (clip(px, 0, map_size - 1) / map_size) - 0.5;
	double y = 0.5 - (clip(py, 0, map_size - 1) / map_size);
	*lat = 90.0 - 360.0 * atan(exp(-y * 2.0 * M_PI)) / M_PI;
	*lon = 360.0 * x;
}


RND_INLINE void bing_tile2quadkey(char *dst, long tilex, long tiley, int zoom)
{
	char *end = dst;
	int n;

	for(n = zoom; n > 0; n--) {
		char digit = '0';
		int mask = 1 << (n - 1);

		if ((tilex & mask) != 0)
			digit++;
		if ((tiley & mask) != 0)
			digit += 2;

		*end = digit;
		end++;
	}
	*end = '\0';
}

#include "cache.c"


/*** Render/draw ***/
static double tile_meter_per_pix[24] = {
	0, 78271.5170, 39135.7585, 19567.8792, 9783.9396, 4891.9698, 2445.9849,
	1222.9925, 611.4962, 305.7481, 152.8741, 76.4370, 38.2185, 19.1093, 9.5546,
	4.7773, 2.3887, 1.1943, 0.5972, 0.2986, 0.1493, 0.0746, 0.0373, 0.0187};

static void img_bing_layer_draw(osm_layer_t *ly, rnd_hid_gc_t gc, const rnd_hid_expose_ctx_t *region, rnd_xform_t *xform_caller)
{
	double lon, lat, t0clon, t0clat;
	int n;
	int level = osm_zoom_rnd2level(region->coord_per_pix);
	char quadkey[64];
	long bpx, bpy;     /* Bing pixel coords (at a specific zoom level) */
	long tile0x, tile0y; /* Bing tile coords (at a specific zoom level) */
	rnd_coord_t t0cx, t0cy; /* left-most Bing tile's center point coords */
	long tilex, tiley;
	rnd_coord_t tcx, tcy;
	time_t now = time(NULL);
	rnd_design_t *dsg = (rnd_design_t *)ly->parent;

	if ((level < 1) || (level > 21))
		return; /* not available */

	if (level > conf_bing.plugins.img_bing.max_zoom_lev)
		level = conf_bing.plugins.img_bing.max_zoom_lev;

	rnd2lldeg(&lon, &lat, region->view.X1, region->view.Y2);

	bing_latlon2xy(lon, lat, level, &bpx, &bpy);
	tile0x = bpx / 256; tile0y = bpy / 256; /* pixel-to-tile */


	/* center of the first tile */
	bpx = tile0x * 256 + 128;
	bpy = tile0y * 256 + 128;
	bing_xy2latlon(bpx, bpy, level, &t0clon, &t0clat);
	lldeg2rnd(t0clon, t0clat, &t0cx, &t0cy);

	t0cx += conf_bing.plugins.img_bing.offs.x;
	t0cy += conf_bing.plugins.img_bing.offs.y;

	rnd_trace("bing draw top-left: %f;%f level=%d\n", lon, lat, level);

	tcx = t0cx; tcy = t0cy;
	tilex = tile0x; tiley = tile0y;

	for(n = 0; n < 64; n++) { /* limit just in case something goes wierd */
		rnd_pixmap_t *pm;
		double size = tile_meter_per_pix[level]*100.0 * 256.0; /* 4.777 is the meter per pixel at this zoom level */
		rnd_coord_t sizec = ceil(size + region->coord_per_pix*0.5); /* round size so there's no gap between tiles */

		bing_tile2quadkey(quadkey, tilex, tiley, level);

		pm = bing_cache_get(dsg, quadkey, now);

/*		rnd_trace("   t0c: %ld;%ld size=%ld region0: %ld;%ld %ld;%ld\n", t0cx, t0cy, size, region->view.X1, region->view.Y1, region->view.X2, region->view.Y2);*/
		if (pm != NULL)
			rnd_render->draw_pixmap(rnd_render, tcx, tcy, sizec, sizec, pm);

		tilex++;
		tcx = floor(tcx + size);
		if (tcx - size/2.0 > region->view.X2) {
			tiley++;
			tcy = floor(tcy - size);
			if (tcy + size/2.0 < region->view.Y1)
				break; /* reached end-of-screen */
			tilex = tile0x;
			tcx = t0cx;
		}
	}
}

static void img_bing_layer_set_offs(osm_layer_t *ly, rnd_coord_t ox, rnd_coord_t oy, int temporary)
{
	if (temporary) {
		*((rnd_coord_t *)&conf_bing.plugins.img_bing.offs.x) = ox;
		*((rnd_coord_t *)&conf_bing.plugins.img_bing.offs.y) = oy;
	}
	else {
		rnd_conf_set_design("plugins/img_bing/offs/x", "%ldnm", (long)ox);
		rnd_conf_set_design("plugins/img_bing/offs/y", "%ldnm", (long)oy);
	}
}

static void img_bing_layer_get_offs(osm_layer_t *ly, rnd_coord_t *ox, rnd_coord_t *oy)
{
	*ox = conf_bing.plugins.img_bing.offs.x;
	*oy = conf_bing.plugins.img_bing.offs.y;
}

/*** Entry points ***/

const char osm_acts_Bing[] = "Bing()";
const char osm_acth_Bing[] = "Create the Bing map image layer\n";
fgw_error_t osm_act_Bing(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	rnd_design_t *hl = RND_ACT_DESIGN;
	osm_design_t *osm = (osm_design_t *)hl;
	osm_layer_t *ly = osm_layer_by_name_bottom(osm, "Bing", 1);

	TODO("get a callback on layer destroy and uninit");
	bing_cache_init(hl);

	ly->plugin_draw = img_bing_layer_draw;
	ly->plugin_set_offs = img_bing_layer_set_offs;
	ly->plugin_get_offs = img_bing_layer_get_offs;
	ly->editable = 0;
	return 0;
}

static void bing_list_iu_ev(rnd_design_t *hidlib, void *user_data, int argc, rnd_event_arg_t argv[])
{
	osm_design_t *osm = (osm_design_t *)hidlib;
	long n;
	int found = 0;

	if ((argc != 3) || (argv[1].type != RND_EVARG_PTR) || (argv[2].type != RND_EVARG_PTR)) {
		rnd_message(RND_MSG_ERROR, "bing_list_iu_ev(): invalid event arguments\n");
		return;
	}

	for(n = 0; n < osm->layers.used; n++) {
		osm_layer_t *ly = osm->layers.array[n];
		if (ly->plugin_draw == img_bing_layer_draw) {
			found = 1;
			break;
		}
	}

	if (found) {
		gds_t *iu = argv[1].d.p, *ioffs = argv[2].d.p;
		if (iu->used > 0)
			gds_append(iu, ';');
		gds_append_str(iu, "Bing");
		if (ioffs->used == 0)
			rnd_append_printf(ioffs, "%.3rm;%.3rm", conf_bing.plugins.img_bing.offs.x, conf_bing.plugins.img_bing.offs.y);
	}
}



static rnd_action_t img_bing_action_list[] = {
	{"bing", osm_act_Bing, osm_acth_Bing, osm_acts_Bing}
};

int pplg_check_ver_img_bing(int ver_needed) { return 0; }

void pplg_uninit_img_bing(void)
{
	rnd_remove_actions_by_cookie(img_bing_cookie);
	rnd_event_unbind_allcookie(img_bing_cookie);
	rnd_conf_plug_unreg("plugins/img_bing/", img_bing_conf_internal, img_bing_cookie);
}

int pplg_init_img_bing(void)
{
	rnd_conf_plug_reg(conf_bing, img_bing_conf_internal, img_bing_cookie);

#define conf_reg(field,isarray,type_name,cpath,cname,desc,flags) \
	rnd_conf_reg_field(conf_bing, field,isarray,type_name,cpath,cname,desc,flags);
#include "img_bing_conf_fields.h"

	RND_REGISTER_ACTIONS(img_bing_action_list, img_bing_cookie);
	rnd_event_bind(OSM_EVENT_LIST_IMAGERY_USED, bing_list_iu_ev, NULL, img_bing_cookie);

	return 0;
}
