/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - simple configurable rendering styles
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */


#include <librnd/config.h>

#include <librnd/core/global_typedefs.h>
#include <librnd/core/hidlib.h>
#include <librnd/core/conf_multi.h>
#include <librnd/core/misc_util.h>
#include <liblihata/dom.h>
#include "event.h"
#include "obj_any.h"
#include "style.h"
#include "std_style_conf.h"

conf_std_style_t conf_std_style;

static const char std_style_cookie[] = "std_style";

#include "conf_internal.c"

RND_INLINE osm_style_t *get_rule_style(lht_node_t *rule)
{
	osm_style_t *st = htsp_get(&osm_name2style, rule->name);
	long id;
	lht_node_t *nd;

	/* get existing */
	if (st != NULL)
		return st;

	/* allocate new */
	id = osm_style_new(rule->name);
	if (id < 0)
		return NULL;
	st = osm_styles.array[id];

	/* fill in fields */
	nd = lht_dom_hash_get(rule, "stroke");
	if ((nd != NULL) && (nd->type == LHT_TEXT)) {
		rnd_color_t clr = {0};
		rnd_color_load_str(&clr, nd->data.text.value);
		st->high.stroke.color = clr;
		st->low.stroke.color = clr;
	}

	nd = lht_dom_hash_get(rule, "high.stroke");
	if ((nd != NULL) && (nd->type == LHT_TEXT))
		rnd_color_load_str(&st->high.stroke.color, nd->data.text.value);

	nd = lht_dom_hash_get(rule, "low.stroke");
	if ((nd != NULL) && (nd->type == LHT_TEXT))
		rnd_color_load_str(&st->low.stroke.color, nd->data.text.value);

	nd = lht_dom_hash_get(rule, "auto_width_meter");
	if ((nd != NULL) && (nd->type == LHT_TEXT)) {
		char *end;
		double w = strtod(nd->data.text.value, &end);
		st->lod_high_below = w;
		st->high.stroke.width = w * 100.0;
		st->low.stroke.width = -w/5;
		if (st->low.stroke.width == 0)
			st->low.stroke.width = -1;
	}

	nd = lht_dom_hash_get(rule, "lod_invis_above");
	if ((nd != NULL) && (nd->type == LHT_TEXT)) {
		char *end;
		st->lod_invis_above = strtol(nd->data.text.value, &end, 10);
	}

	nd = lht_dom_hash_get(rule, "zord");
	if ((nd != NULL) && (nd->type == LHT_TEXT)) {
		char *end;
		st->zord = strtol(nd->data.text.value, &end, 10);
	}

	nd = lht_dom_hash_get(rule, "area");
	if ((nd != NULL) && (nd->type == LHT_TEXT))
		st->area = rnd_istrue(nd->data.text.value);

	return st;
}

RND_INLINE void std_style_determine(osm_style_t **res, osm_any_obj_t *obj, osm_tags_t *tags, const rnd_conflist_t *cfg)
{
	gdl_iterator_t it;
	rnd_conf_listitem_t *i;

	rnd_conflist_foreach(cfg, &it, i) {
		lht_node_t *tag, *val, *rule = i->prop.src;
		char *tv;

		if (rule->type != LHT_HASH) continue;
		tag = lht_dom_hash_get(rule, "tag");
		val = lht_dom_hash_get(rule, "val");
		if ((tag == NULL) || (val == NULL))
			continue;

		tv = htsp_get(tags, tag->data.text.value);
		if (tv == NULL)
			continue;
		if (strcmp(tv, val->data.text.value) == 0) {
			osm_style_t *style = get_rule_style(rule);
			if (style != NULL) {
				*res = style;
				return;
			}
		}
	}
}

RND_INLINE void std_style_by_name(osm_style_t **res, osm_any_obj_t *obj, const char *name, const rnd_conflist_t *cfg)
{
	gdl_iterator_t it;
	rnd_conf_listitem_t *i;

	rnd_conflist_foreach(cfg, &it, i) {
		lht_node_t *rule = i->prop.src;
		if (strcmp(rule->name, name) == 0) {
			*res = get_rule_style(rule);
			return;
		}
	}
}

static void std_style_determine_ev(rnd_design_t *hidlib, void *user_data, int argc, rnd_event_arg_t argv[])
{
	osm_style_t **res;
	osm_any_obj_t *obj;
	osm_tags_t *tags = NULL;
	osm_layer_t *ly;
	const rnd_conflist_t *cfg;
	int lyforce;

	if ((argc != 3) || (argv[1].type != RND_EVARG_PTR) || (argv[2].type != RND_EVARG_PTR))
		return;

	res = argv[1].d.p;
	obj = argv[2].d.p;
	ly = obj->hdr.layer;
	lyforce = (ly->force_style_name != NULL);


	if (obj->hdr.calls->get_tags != NULL)
		tags = obj->hdr.calls->get_tags(obj);

	if ((tags == NULL) && !lyforce)
		return;

	switch(obj->hdr.type) {
		case OSM_OBJ_NODE: cfg = &conf_std_style.plugins.std_style.nodes; break;
		case OSM_OBJ_WAY:  cfg = &conf_std_style.plugins.std_style.ways; break;
		case OSM_OBJ_RELATION:
		case OSM_OBJ_invalid:
		case OSM_OBJ_max:
			return;
	}

	if (lyforce)
		std_style_by_name(res, obj, ly->force_style_name, cfg);
	else
		std_style_determine(res, obj, tags, cfg);
}



int pplg_check_ver_std_style(int ver_needed) { return 0; }

void pplg_uninit_std_style(void)
{
	rnd_conf_plug_unreg("plugins/std_style/", std_style_conf_internal, std_style_cookie);
	rnd_event_unbind_allcookie(std_style_cookie);
}

int pplg_init_std_style(void)
{
	rnd_conf_plug_reg(conf_std_style, std_style_conf_internal, std_style_cookie);

#define conf_reg(field,isarray,type_name,cpath,cname,desc,flags) \
	rnd_conf_reg_field(conf_std_style, field,isarray,type_name,cpath,cname,desc,flags);
#include "std_style_conf_fields.h"

	rnd_event_bind(OSM_EVENT_STYLE_DETERMINE, std_style_determine_ev, NULL, std_style_cookie);

	return 0;
}
