/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - standard tools
 *  Copyright (C) 2023,2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "crosshair.h"
#include "conf_core.h"
#include "data.h"
#include "obj_any.h"
#include "search.h"
#include "select.h"
#include <librnd/core/compat_misc.h>
#include <librnd/hid/hid.h>
#include <librnd/hid/hid_inlines.h>


typedef struct {
	osm_layer_t *ly;
	rnd_coord_t press_x, press_y, adjx, adjy;
	int click;
	int copy;
	int ret_arrow;
	int endpoint; /* 1 or 2 for grabbing endpoint in move, 0 for non-endpoint */
	vtp0_t click_list;
	vtp0_t draw_list; /* also draw affected objects, e.g. a node of a way being moved should trigger movedraw on the way but not move of the way */
} tool_movecopy_t;

static tool_movecopy_t mvcp;

static void tool_move_init(void)
{
	mvcp.copy = 0;
}

static void tool_copy_init(void)
{
	mvcp.copy = 1;
}

static void tool_movecopy_uninit(void)
{
}

/* for the passon from the arrow tool */
static void tool_movecopy_press_at(osm_layer_t *ly, rnd_coord_t x, rnd_coord_t y, osm_any_obj_t *obj_in, int ret_arrow)
{
	long v, n;
	static const osm_objtype_mask_t draw_list_mask = OSM_MASK_WAY;

	if (ly == NULL) {
		rnd_message(RND_MSG_ERROR, "Select a layer first!\n");
		return;
	}

	if (!ly->editable) {
		rnd_message(RND_MSG_ERROR, "Can't move object: layer is not editable\n");
		if (ret_arrow)
			rnd_tool_select_by_name(&ly->parent->hidlib, "arrow");
		return;
	}

	mvcp.ly = ly;
	mvcp.ret_arrow = ret_arrow;
	mvcp.press_x = x;
	mvcp.press_y = y;
	mvcp.adjx = x;
	mvcp.adjy = y;

	mvcp.draw_list.used = 0;
	mvcp.click_list.used = 0;
	mvcp.endpoint = 0;
	v = osm_search_all_selected(&mvcp.click_list, ly);

	/* if there's no selection, go with single object, either from obj_in or
	   if that's NULL by a screen search */
	if (v == 0) {
		if (obj_in == NULL)
			obj_in = osm_search_obj_xy(ly, mvcp.press_x, mvcp.press_y);
		if (obj_in == NULL) {
			mvcp.click = 0;
			return;
		}
		vtp0_append(&mvcp.click_list, obj_in);
	}

	/* mark all nodes referers so they are not added again when we collect
	   referee nodes for clicked objects */
	for(n = 0; n < mvcp.click_list.used; n++) {
		osm_any_obj_t *obj = mvcp.click_list.array[n];
		if (obj->hdr.type == OSM_OBJ_NODE)
			obj->hdr.referer = 1;
	}

	/* figure what affected objects are to be drawn as well; e.g. if a node
	   of a way is moved, the way need to be redrawn as well */
	for(n = 0; n < mvcp.click_list.used; n++) {
		osm_any_obj_t *obj = mvcp.click_list.array[n];
		osm_obj_list_referers(&mvcp.draw_list, obj, draw_list_mask);
		/* plus a complex object (e.g. a way) is moved by moving all its nodes,
		   so pretend those were clicked too (without selecting them) */
		osm_obj_referee_nodes(obj, &mvcp.click_list, 1);
	}

	/* finally update the referee node attached coords of any object we are
	   ever going to draw-attached */
	for(n = 0; n < mvcp.draw_list.used; n++) {
		osm_any_obj_t *obj = mvcp.draw_list.array[n];
		osm_obj_referee_nodes(obj, NULL, 1);
	}

	osm_obj_list_clear_referer(&mvcp.draw_list);
	osm_obj_list_clear_referer(&mvcp.click_list);

	mvcp.click = 1;
}

static void tool_movecopy_press(rnd_design_t *hl)
{
	osm_design_t *osm = (osm_design_t *)hl;
	osm_layer_t *ly = osm_layer_get_curr(osm);

	tool_movecopy_press_at(ly, osm->crosshair_x, osm->crosshair_y, NULL, 0);
}

static void tool_movecopy_release(rnd_design_t *hl)
{
	long n;

	if (mvcp.click) {
		osm_layer_t *layer = mvcp.ly;

		if (!mvcp.copy) {
			if (!mvcp.endpoint) {
				uundo_freeze_serial(&layer->undo);

				/* move nodes to their last known attached-coords (only nodes have coordinates) */
				for(n = 0; n < mvcp.click_list.used; n++) {
					osm_any_obj_t *obj = mvcp.click_list.array[n];
					if (obj->hdr.type == OSM_OBJ_NODE)
						osm_node_move_to_att(&obj->node, 1);
				}

				/* recalculate the bbox of all affected objects */
				for(n = 0; n < mvcp.draw_list.used; n++) {
					osm_any_obj_t *obj = mvcp.draw_list.array[n];
					osm_obj_bbox_update(obj);
				}

				uundo_unfreeze_serial(&layer->undo);
				uundo_inc_serial(&layer->undo);
			}
		}
		else {
			uundo_freeze_serial(&layer->undo);
/*
			for(n = 0; n < mvcp.click_list.used; n++)
				osm_obj_copy(mvcp.click_list.array[n], dx, dy, 1);
*/
			uundo_unfreeze_serial(&layer->undo);
			uundo_inc_serial(&layer->undo);
		}

		mvcp.click = 0;
	}
	if (mvcp.ret_arrow)
		rnd_tool_select_by_name(hl, "arrow");
}

static void tool_movecopy_adjust_attached_objects(rnd_design_t *hl)
{
	osm_design_t *osm = (osm_design_t *)hl;
	mvcp.adjx = osm->crosshair_x;
	mvcp.adjy = osm->crosshair_y;
}

RND_INLINE void tool_movecopy_draw_att(osm_any_obj_t *o, rnd_coord_t dx, rnd_coord_t dy, rnd_hid_gc_t gc)
{
	if (o->hdr.type == OSM_OBJ_NODE) {
		o->node.att_x = o->node.x + dx;
		o->node.att_y = o->node.y + dy;
	}
	osm_obj_draw_attached(o, osm_crosshair_gc);
}


static void tool_movecopy_draw_attached(rnd_design_t *hl)
{
	if (mvcp.click) {
		rnd_coord_t x = mvcp.adjx, y = mvcp.adjy;
		rnd_coord_t dx = x - mvcp.press_x, dy = y - mvcp.press_y;
		long n;

		rnd_hid_set_line_cap(osm_crosshair_gc, rnd_cap_round);
		rnd_hid_set_line_width(osm_crosshair_gc, -1);

		/* draw these translated by dx;dy */
		for(n = 0; n < mvcp.click_list.used; n++)
			tool_movecopy_draw_att(mvcp.click_list.array[n], dx, dy, osm_crosshair_gc);

		/* draw these original */
		for(n = 0; n < mvcp.draw_list.used; n++)
			osm_obj_draw_attached(mvcp.draw_list.array[n], osm_crosshair_gc);
	}
}

/* XPM */
static const char *move_icon[] = {
/* columns rows colors chars-per-pixel */
"21 21 3 1",
"  c #000000",
". c #6EA5D7",
"o c None",
/* pixels */
"ooooooooooooooooooooo",
"ooooooooooooooooooooo",
"oooooooooooo  ooooooo",
"oooooooooooo . oooooo",
"ooo          .. ooooo",
"ooo ............ oooo",
"ooo ............ oooo",
"ooo          .. ooooo",
"oooooooooooo . oooooo",
"oooooooooooo  ooooooo",
"ooooooooooooooooooooo",
"ooooooooooooooooooooo",
"ooooooooooooooooooooo",
"o ooo oo   ooo ooo oo",
"o  o  o ooo oo ooo oo",
"o     o ooo oo ooo oo",
"o o o o ooo oo ooo oo",
"o ooo o ooo oo ooo oo",
"o ooo o ooo ooo o ooo",
"o ooo o ooo ooo o ooo",
"o ooo oo   oooo   ooo"
};

/* XPM */
static const char *copy_icon[] = {
/* columns rows colors chars-per-pixel */
"21 21 3 1",
"  c #000000",
". c #6EA5D7",
"o c None",
/* pixels */
"ooooooooooooooooooooo",
"ooooooooooooooooooooo",
"oooooo  oooooo   oooo",
"oooooo . ooooo .. ooo",
"o      .. ooo  ..  oo",
"o ........ o ...... o",
"o ........ o ...... o",
"o      .. ooo  ..  oo",
"oooooo . ooooo .. ooo",
"oooooo  oooooo    ooo",
"ooooooooooooooooooooo",
"ooooooooooooooooooooo",
"ooooooooooooooooooooo",
"o     o    ooo ooo oo",
"o  oooo ooo oo ooo oo",
"o ooooo ooo ooo o ooo",
"o ooooo    oooo   ooo",
"o ooooo oooooooo oooo",
"o ooooo oooooooo oooo",
"o ooooo oooooooo oooo",
"o     o oooooooo oooo"
};


static rnd_tool_t osm_rnd_tool_move = {
	"move", NULL, std_tools_cookie, 50, move_icon, RND_TOOL_CURSOR_NAMED("left_ptr"), 0,
	tool_move_init,
	tool_movecopy_uninit,
	tool_movecopy_press,
	tool_movecopy_release,
	tool_movecopy_adjust_attached_objects,
	tool_movecopy_draw_attached,
	NULL, /* undo */
	NULL, /* redo */
	NULL, /* escape */
};

static rnd_tool_t osm_rnd_tool_copy = {
	"copy", NULL, std_tools_cookie, 50, copy_icon, RND_TOOL_CURSOR_NAMED("left_ptr"), 0,
	tool_copy_init,
	tool_movecopy_uninit,
	tool_movecopy_press,
	tool_movecopy_release,
	tool_movecopy_adjust_attached_objects,
	tool_movecopy_draw_attached,
	NULL, /* undo */
	NULL, /* redo */
	NULL, /* escape */
};
