/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - standard tools
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"

#include <genvector/vtp0.h>
#include <genvector/vtl0.h>

#include "data.h"
#include "crosshair.h"
#include "projection.h"
#include <librnd/hid/tool.h>

static vtp0_t ntmp;
static vtl0_t ltmp;

void osm_tool_node_uninit(void)
{
	vtp0_uninit(&ntmp);
	vtl0_uninit(&ltmp);
}

osm_node_t *osm_tool_node_insert(osm_layer_t *ly)
{
	osm_design_t *osm = ly->parent;
	double lon, lat;
	osm_any_obj_t *obj;
	int inserted = 0;

	uundo_freeze_serial(&ly->undo);

	/* if clicked on a way and ctrl is not pressed, insert in way*/
	if (!rnd_gui->control_is_pressed(rnd_gui)) {
		rnd_coord_t maxdist2 = osm_slop * osm_slop * 10;
		ntmp.used = ltmp.used = 0;
		osm_search_append_objs_xy_mask(&ntmp, ly, osm->crosshair_x, osm->crosshair_y, OSM_MASK_WAY);

		if (ntmp.used == 1) {
			osm_any_obj_t *way = ntmp.array[0];
			obj = (osm_any_obj_t *)osm_way_modify_insert_near(&way->way, osm->crosshair_x, osm->crosshair_y, maxdist2, 1);
			inserted = 1;
		}
		else if (ntmp.used > 1) {
			long n, idx, numpts = 0;
			double px = 0, py = 0;
			rnd_coord_t rx, ry;

			/* calculate center of points */
			for(n = 0; n < ntmp.used; n++) {
				osm_any_obj_t *way = ntmp.array[n];
				idx = osm_way_find_near(&way->way, osm->crosshair_x, osm->crosshair_y, maxdist2, &rx, &ry);
				if (idx >= 0) {
					px += rx;
					py += ry;
					numpts++;
				}
				vtl0_append(&ltmp, idx);
			}

			rnd_trace("cross! numpts=%ld\n", numpts);

			/* create the pointinsert in each way */

			if (numpts > 0) {
				px /= numpts;
				py /= numpts;

				obj = osm_obj_create(ly, OSM_OBJ_NODE, 1);
				rnd2lldeg(&lon, &lat, rnd_round(px), rnd_round(py));
				osm_node_modify(ly, &obj->node, &lon, &lat, 1);
				inserted = 1;

				for(n = 0; n < ntmp.used; n++) {
					osm_any_obj_t *way = ntmp.array[n];
					idx = ltmp.array[n];
					if (idx >= 0)
						osm_way_modify_insert_before(ly, &way->way, obj->hdr.id, idx, 1);
				}
			}
		}
	}

	if (!inserted) {
		obj = osm_obj_create(ly, OSM_OBJ_NODE, 1);
		rnd2lldeg(&lon, &lat, osm->crosshair_x, osm->crosshair_y);
		osm_node_modify(ly, &obj->node, &lon, &lat, 1);
	}

	uundo_unfreeze_serial(&ly->undo);
	uundo_inc_serial(&ly->undo);

	return obj == NULL ? NULL : &obj->node;
}

void osm_tool_node_press(rnd_design_t *hl)
{
	osm_design_t *osm = (osm_design_t *)hl;
	osm_layer_t *ly = osm_layer_get_curr(osm);

	if (ly == NULL) {
		rnd_message(RND_MSG_ERROR, "Select a layer first!\n");
		return;
	}

	if (!ly->editable) {
		rnd_message(RND_MSG_ERROR, "Can't create node: layer is not editable\n");
		return;
	}

	osm_tool_node_insert(ly);
}

void osm_tool_node_adjust_attached_objects(rnd_design_t *hl)
{
}

void osm_tool_node_draw_attached(rnd_design_t *hl)
{
}

/* XPM */
static const char *node_icon[] = {
/* columns rows colors chars-per-pixel */
"21 21 3 1",
"  c #000000",
". c #6EA5D7",
"o c None",
/* pixels */
"ooooooooooooooooooooo",
"ooooooooooooooooooooo",
"ooooooooooooooooooooo",
"ooooooo       ooooooo",
"ooooooo ..... ooooooo",
"ooooooo .ooo. ooooooo",
"ooooooo .ooo. ooooooo",
"ooooooo .ooo. ooooooo",
"ooooooo ..... ooooooo",
"ooooooo       ooooooo",
"ooooooooooooooooooooo",
"ooooooooooooooooooooo",
"ooooooooooooooooooooo",
"o oooo oo  oo o oo   ",
"o  ooo o oo o oo o oo",
"o o oo o oo o oo o oo",
"o o oo o oo o oo o  o",
"o oo o o oo o oo o oo",
"o oo o o oo o oo o oo",
"o ooo  o oo o oo o oo",
"o oooo oo  oo   oo   "
};

rnd_tool_t osm_rnd_tool_node = {
	"node", NULL, NULL, 100, node_icon, RND_TOOL_CURSOR_NAMED("up_arrow"), 0,
	NULL,
	osm_tool_node_uninit,
	osm_tool_node_press,
	NULL,
	osm_tool_node_adjust_attached_objects,
	osm_tool_node_draw_attached,
	NULL,
	NULL,
	NULL, /* escape */
};
